import { app, BrowserWindow, screen, ipcMain, IpcMainEvent, shell } from 'electron';
import * as path from 'path';
import * as url from 'url';
import * as fs from 'fs';
import * as YAML from 'yaml';
import * as child from 'promisify-child-process';
import * as util from 'util';
import * as temp from 'tmp';
import { Job } from './src/app/job';
import { AppAction, Strudelapp, StrudelappInstance } from './src/app/strudelapp';
import * as net from "net";
import { stdout } from 'process';
import { SpawnOptionsWithStdioTuple } from 'child_process';
import { setLines, ShowOnDirtyErrorStateMatcher } from '@angular/material/core';
import { SELECT_PANEL_INDENT_PADDING_X } from '@angular/material/select/select';
import { window } from 'rxjs-compat/operator/window';

let win: BrowserWindow = null;
const args = process.argv.slice(1),
  serve = args.some(val => val === '--serve');
let lastUrl: string = null;
console.log('resources path', __dirname );
var config: any;
var sshOpts = [ '-o', 'StrictHostKeyChecking=no',
'-o', 'UserKnownHostsFile=/dev/null',
'-o', 'IdentityFile=/dev/null',
'-o', 'PreferredAuthentications=publickey',
'-o', 'UpdateHostKeys=no',
'-o', 'LogLevel=error'];



let configPath = path.join(app.getPath('userData'),'localConfig.yml');
try {
  config = YAML.parse(fs.readFileSync(configPath).toString());
} catch {
  console.log(`please create a config file at ${configPath}`)

}
let testconfig =YAML.parse(fs.readFileSync(path.join(__dirname,'resources','testConfig.yml')).toString()); 
console.log(testconfig);
//let config = YAML.parse(fs.readFileSync(path.join(__dirname,'localConfig.yml')).toString())
//var aafds: string = "https://ds.aaf.edu.au/discovery/aaf/9DrjGhkyXjXsPm45rqpCXA?entityID=https%3A%2F%2Fcentral.aaf.edu.au%2Fshibboleth&return=https%3A%2F%2Fcentral.aaf.edu.au%2FShibboleth.sso%2FLogin%3FSAMLDS%3D1%26identity%3D1%26target%3Dhttps%253A%252F%252Fcentral.aaf.edu.au%252Fresolvers%252Fsaml_federation%252Freceive%253Fibr%253D19a58192-de47-4580-a272-cc61c60e4f47%2526pib%253D0a206a4e-e6eb-4d18-b46d-e475ec8d09b5"
function createWindow(): BrowserWindow {

  const electronScreen = screen;
  const size = electronScreen.getPrimaryDisplay().workAreaSize;

  // Create the browser window.
  const win = new BrowserWindow({
    x: 0,
    y: 0,
    width: size.width,
    height: size.height,
    webPreferences: {
      nativeWindowOpen: false,
      nodeIntegration: false,
      allowRunningInsecureContent: (serve) ? true : false,
      contextIsolation: true,  // false if you want to run 2e2 test with Spectron
      //enableRemoteModule : false, // true if you want to run 2e2 test  with Spectron or use remote module in renderer context (ie. Angular)
      preload: path.join(__dirname, "preload.js")
    },
  });
  win.webContents.setWindowOpenHandler((details) => {
    if (details.url.startsWith('http://localhost:')) {
      return { action: 'allow',
                overrideBrowserWindowOptions: {
                disposition: 'foreground-tab'
              }
      }
    
      //shell.openExternal(url);
      //return { action: 'deny' };
    }
    if (details.url.startsWith('vnc://localhost:')) {
      return { action: 'allow' }
    }
    return { action: 'deny' }
  })


  if (serve) {

    win.webContents.openDevTools();

    require('electron-reload')(__dirname, {
      electron: require(`${__dirname}/node_modules/electron`)
    });
    win.loadURL(url.format({
      pathname: path.join(__dirname, 'dist/index.html'),
      protocol: 'file:',
      slashes: true
    }));
    //win.loadURL('http://localhost:4200');

  } else {
    win.loadURL(url.format({
      pathname: path.join(__dirname, 'dist/index.html'),
      protocol: 'file:',
      slashes: true
    }));
  }

  // Emitted when the window is closed.

  win.on('closed', () => {
    // Dereference the window object, usually you would store window
    // in an array if your app supports multi windows, this is the time
    // when you should delete the corresponding element.
    //win = null;
  });

  win.webContents.on('did-fail-load', () => {
    win.loadURL(url.format({
      pathname: path.join(__dirname, 'dist/index.html'),
      protocol: 'file:',
      slashes: true
    }));
    // REDIRECT TO FIRST WEBPAGE AGAIN
      });

  const redirectUri = 'http://localhost:4200/sshauthz_callback'
  const filter = {
     urls: [redirectUri + '*']
  };
  win.webContents.session.webRequest.onBeforeRequest(filter, function(details, callback) {
   
      win.loadURL(url.format({
        pathname: path.join(__dirname, 'dist/index.html'),
        protocol: 'file:',
        slashes: true
      }));

      win.webContents.on('did-finish-load', () => {
        if (lastUrl != details.url) {
          win.webContents.send('oauth2-redirect',details.url);
          lastUrl = details.url;
        }
      })
      callback({ cancel: false });
  });


  return win;
}

try {
  // This method will be called when Electron has finished
  // initialization and is ready to create browser windows.
  // Some APIs can only be used after this event occurs.
  // Added 400 ms to fix the black background issue while using transparent window. More detais at https://github.com/electron/electron/issues/15947
  app.on('ready', () => setTimeout(createWindow, 400));

  // Quit when all windows are closed.
  app.on('window-all-closed', () => {
    // On OS X it is common for applications and their menu bar
    // to stay active until the user quits explicitly with Cmd + Q
    if (process.platform !== 'darwin') {
      app.quit();
    }
  });
  ipcMain.on('addCert', async (event: IpcMainEvent, data: any) => await addCert(event, data));
  ipcMain.on('launch', async (event: IpcMainEvent, job: Job, appinst: StrudelappInstance, action: AppAction) => await launch(event, job, appinst, action));
  ipcMain.on('createTunnel', async (event: IpcMainEvent, data: any) => await createTunnel(event, data));
  ipcMain.on('runRemoteCommand', async (event: IpcMainEvent, data: any) => await runRemoteCommand(event, data));
  ipcMain.on('sshAgent', async (event: IpcMainEvent, data: any)  => await get_agent_contents(event, data));




  app.on('activate', () => {
    // On OS X it's common to re-create a window in the app when the
    // dock icon is clicked and there are no other windows open.
    if (win === null) {
      createWindow();

    }
  });
  

} catch (e) {
  // Catch Error
  // throw e;
}

async function addCert(event: IpcMainEvent, data: any) {

  let tmpfile = temp.fileSync();
  let path = tmpfile.name;
  fs.writeFileSync(`${path}`,data.key, {encoding: 'utf8', mode: 0o600 });
  fs.writeFileSync(`${path}-cert.pub`,data.cert, {encoding: 'utf8', mode: 0o600 });
  let re = new RegExp('Valid: from ([0-9:T-]*) to ([0-9:T-]*)')
  let key_gen = child.spawn(`${config['ssh-keygen']}`,['-L','-f',`${path}-cert.pub`],{encoding: 'utf8' });
  try {
    var kg: string = <string>(await key_gen).stdout;
    let match = re.exec(kg);
    let now = new Date();
    let expiry = new Date(match[2]); // ssh-keygen reports expiry in the local timezone and new Date assumes the parameters are in the local timezone
    let seconds_remaining = Math.floor((expiry.getTime()-now.getTime())/1000);
    let ssh_add = child.spawn(`${config['ssh-add']}`,['-t',`${seconds_remaining}`,`${path}`]);
    await ssh_add;
    event.reply('addCertResponse',null);
  } catch (error) {
  }
  fs.unlinkSync(`${path}`);
  fs.unlinkSync(`${path}-cert.pub`);

}

function parse_cert(u: string ) {
  var key = null;
  var res = {};
  var values = [];
  var v;
  for (let l of u.split("\n")) {
      var tl = l.trimLeft().trimRight();
      if (tl.includes(":")) {
          if (key !== null) {
              res[key] = values;
          }
          values = []
          try {
            let a = tl.split(':');
            key = a[0];
            v = a.slice(1).join(':');
            let vt = v.trimLeft().trimRight();
            if (vt != "") {
                values = [vt]
            }
          } catch {
          }

      } else {
          if (l != "") {
              values.push(l.trimLeft().trimRight())
          }
      }
  }
  return res;
}

async function get_agent_contents(event: IpcMainEvent, data: any) {
  try {
    var ssh_add = child.spawn(`ssh-add`,['-L'], {encoding: 'utf8'});
    var agent_lines: string = <string>(await ssh_add).stdout;
  } catch (error) {
    agent_lines = "";
  }
  var r = []
  for (let l of agent_lines.split("\n")) {
      try {
        var ssh_keygen = child.spawn('ssh-keygen',['-L','-f','-'], {encoding: 'utf8'})
        ssh_keygen.stdin.write(l);
        ssh_keygen.stdin.end();
        var stdout: string = <string>(await ssh_keygen).stdout;
        if (stdout!== undefined) {
            r.push(parse_cert(<string>stdout));
        }
      } catch (error) {
      }
  }
  event.reply('sshAgentResponse', null,r,null);
}


async function launch(event: IpcMainEvent, job: Job, appinst: StrudelappInstance, action: AppAction) {
  let tag = 'launchResponse'
  console.log('in main.ts launch', job, appinst, action)
  var cmd = 'echo';
  let args = ['launch not implemented'];
  //wrap_execute( cmd, args, null, (e,so,se)  => wrap_execute_callback(event, tag, e,so,se));
}

function sleep(ms) {
  return new Promise(resolve => setTimeout(resolve, ms));
}
async function createTunnelWithPort(event: IpcMainEvent, data: any, localPort: any) {
  console.log('calling CreateTunnelWithPort '+localPort);
  let remotePort = data['appinst']['port']
  var cmd = `${config['ssh']}`
  var args= sshOpts.concat(['-f', `-L`, `${localPort}:localhost:${remotePort}`, `-l`, `${data['user']}`, `-o`, `ProxyJump=${data['bastion']}`, `${data['host']}`, 'sleep','60']);
  let cp = child.spawn(cmd, args);
  await sleep(2000);
  event.reply('createTunnelResponse',null,{'localport':localPort},null);
}

async function createTunnel(event: IpcMainEvent, data: any) {
  //let tag = 'createTunnelResponse'
  var srv = net.createServer(function(sock) {
    sock.end('Hello world\n');
  });

  srv.listen({'port': 0});
  srv.on('listening', () => { 
    var port = srv.address()['port'] ; 
    srv.close();
    createTunnelWithPort(event, data, port)
  })
  // I'm worried about this. I think srv.listen is async, so I should be using a promise to wait for it to get a port and be ready to proceed
  //var localPort = srv.address()['port']
  //srv.close();



  //wrap_execute( cmd, (e,so,se)  => wrap_execute_callback(event, tag, e,so,se), null);
}

async function killPid(event: IpcMainEvent, data: any) {
}

async function runRemoteCommand(event: IpcMainEvent, data: any) {
  var tag;
  if (data.hasOwnProperty('tag')) {
    tag = data['tag']
  } else {
    tag = 'runRemoteCommandResponse'+data['cmd'];
  }
  var cmd;
  var args;
  if (data.hasOwnProperty('bastion')) {
    cmd = `${config['ssh']}`;
    args = sshOpts.concat([`-l`, `${data['user']}`, `-o`, `ProxyJump=${data['bastion']}`, `${data['host']}`, `${data['cmd']}`]);
  } else {
    cmd = `${config['ssh']}`;
    args = sshOpts.concat([`-l`,`${data['user']}`, `${data['host']}`,`${data['cmd']}`]);
  }

  var rc = child.spawn(cmd, args, {encoding: 'utf8', 'timeout':60000})
  if (data.hasOwnProperty('input')) {
    rc.stdin.write(data['input']);
    rc.stdin.end();
  }
  try {
    var rc_out = await rc;
    event.reply(tag, null, rc_out.stdout, rc_out.stderr);
  } catch (error) {
    console.log('error on',cmd,args);
    event.reply(tag, error.code, error.stdout, error.stderr);
  }
  //wrap_execute( cmd, args, stdin, (e,so,se)  => wrap_execute_callback(event, tag, e,so,se));

}